# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Built-in rules determining what DDD sequence should be applied in a given slack window."""
from mitiq.ddd.rules.rules import xx, xyxy, yy, general_rule, repeated_rule
